"""
    These are the sub-commands for a top level object added to the CLI group in 'cli.py'.
    The commands and options are implemented here and the logic behind them resides in the corresponding behavior file.
"""

from cli import pyke, behavior
import click

@click.group('component-set', help='A component-set is comprised of component-set-versions. You can share component-sets with other orgs giving them access to use the associated versions.')
def component_set():
    pass

@click.command('ls')
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command.', autocompletion=pyke.auth.get_profile_names)
@click.option('--format', '-f', default='{id} {name} {urlRef} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--filter', default='')
@click.option('--page', default=1)
@click.option('--pagesize', default=100)
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
@pyke.auth.profile_context_required
def list(format, profile, json, **kwargs):
    behavior.ComponentSetBehavior(profile=profile, format=format, json=json, **kwargs).list()

@click.command('rm')
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command.', autocompletion=pyke.auth.get_profile_names)
@click.option('--component-set', '-cs', prompt=True, help='The ID or the urlRef of the component-set to delete')
@click.option('--format', '-f', default='{id} {name} {urlRef} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
@pyke.auth.profile_context_required
def delete(component_set, profile, format, json):
    behavior.ComponentSetBehavior(profile=profile, format=format, json=json).delete(component_set)

@click.command('add')
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command.', autocompletion=pyke.auth.get_profile_names)
@click.option('--name', prompt=True, help='The name of the component set')
@click.option('--url-ref', prompt=True, help="A unique value for how the component-set will be addressed in a URL.")
@click.option('--icon-file', '-path', type=click.Path(), help='The path to an SVG icon for the component set')
@click.option('--format', '-f', default='{id} {name} {urlRef} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
@pyke.auth.profile_context_required
def create(profile, name, url_ref, icon_file, format, json):
    behavior.ComponentSetBehavior(profile=profile, format=format, json=json).create(name, url_ref, icon_file)

@click.command()
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command.', autocompletion=pyke.auth.get_profile_names)
@click.option('--component-set', '-cs', prompt=True, help='The ID or the urlRef of the component-set to update')
@click.option('--name', default='', help='The name of the component set')
@click.option('--icon-file', '-path', default='', type=click.Path(), help='The path to an SVG icon for the component set')
@click.option('--format', '-f', default='{id} {name} {urlRef} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
@pyke.auth.profile_context_required
def update(component_set, name, icon_file, profile, format, json):
    behavior.ComponentSetBehavior(profile=profile, format=format, json=json).update(component_set, name, icon_file)

@click.command('access', help="The access command shares and unshares objects with child orgs.")
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command.', autocompletion=pyke.auth.get_profile_names)
@click.option('--component-set', '-cs', prompt=True, help='The id or the urlRef of the component-set being shared/unshared')
@click.option('--add', multiple=True, help="Share with an org. (ID || Org Name)")
@click.option('--rm', multiple=True, help="Un-Share with an org. (ID || Org Name)")
@click.option('--absolute', multiple=True, help='Flush and fill shared with list. (ID || Org Name)')
@click.option('--format', '-f', default='{failed} {sharedWith} {unsharedFrom} {resultingGrantees}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
@click.option('--current', is_flag=True, help="Returns a list of orgs currently shared with this object.")
@pyke.auth.profile_context_required
def share(component_set, add, rm, absolute, profile, format, json, current):
    behavior.ComponentSetBehavior(profile=profile, format=format, json=json).share(component_set, add, rm, absolute, current)

component_set.add_command(list)
component_set.add_command(create)
component_set.add_command(delete)
component_set.add_command(update)
component_set.add_command(share)



from .component_set_version import *
from .component_set import *
from .organization import *
from .env import *
from .widget import *
from .widget_version import *
from .profile import *
from .microservice import *
from .microservice_version import *
from .api import *
from .experience import *
from .experience_collection import *
from .container_version import *
from .project import *



"""
    The Util class is comprised of common code that is used throughout the CLI.
    Refreshing tokens, making calls to the platform, building urls, and looking up
    various lumavate objects are few things implamented here.

    TO DO: This needs to be broken up further. Refreshing tokens and all other things
    auth should be broken out into the auth class. Request handling and the various validation
    of objects sprinkled in this code and the behavior files are two other major chunks of code
    that need a new home.
"""

from pathlib import Path
from pyparsing import *
import dateutil.parser
from cli import pyke
import requests
import platform
import urllib3
import urllib
import click
import copy
import json
import sys
import os


ESC = Literal('\x1b')
integer = Word(nums)
escapeSeq = Combine(ESC + '[' + Optional(delimitedList(integer,';')) +
                oneOf(list(alphas)))

uncolor = lambda s : Suppress(escapeSeq).transformString(s)

verify_tls = os.environ.get('LUMA_CLI_VERIFY_TLS', None)
if verify_tls == 'false' or verify_tls == 'False':
  verify_tls = False
  urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
else:
  verify_tls = None

class Util:
  def __init__(self, context=None, profile=None):
    self.context = pyke.auth.load_cache() if not context else context
    if profile is None:
      self.profile = self.context.get('profileName')
    else:
      self.profile = profile

  def login_token(self, renew=False):
    if renew:
      self.context['accessToken'] = self.login()['access_token']
      return self.context['accessToken']

    if self.context is not None:
      return self.context.get('accessToken')

  def login_without_context(self, env, config_data=None):
    if config_data is None:
      config_data = pyke.auth.load_config()
    else:
      config_data = config_data

    env = config_data['envs'][env]

    data = {
      'grant_type': 'client_credentials',
      'audience': env['audience'],
      'client_id': env['clientId'],
      'client_secret': env['clientSecret']
    }

    req_data = pyke.auth.sanatize_data(data)

    token_url = env['token'].replace(" ", "")
    resp = self.cli_request('POST', 'https://' + token_url+ '/oauth/token', json=req_data, external=True)

    if resp.status_code != 200:
      click.echo(json.dumps(env))
      click.echo(' ')

      raise click.ClickException(click.style("Env not configured correctly. Make sure your Client ID and Client Secret are correct.", fg='red'))

    return resp.json()['access_token']

  def login(self):
    data = {
      'grant_type': 'client_credentials',
      'audience': self.context['audience'].replace(" ", ""),
      'client_id': self.context['clientId'].replace(" ", ""),
      'client_secret': self.context['clientSecret'].replace(" ", "")
    }

    res = self.cli_request('POST', 'https://{}/oauth/token'.format(self.context['token']), json=data, external=True)

    if res.status_code != 200:
      raise click.ClickException(click.style("Env not configured correctly. Make sure your Client ID and Client Secret are correct.", fg='red'))

    resp = res.json()

    if self.context.get('orgId') is not None:
      url = '{}/auth/v1/session'.format(self.context.get('app'))
      data = {'companyId': self.context.get('orgId')}
      r = requests.post(url, json=data, headers={'Authorization': 'Bearer ' + resp['access_token']}, verify=verify_tls)
      if r.status_code != 200:
        raise click.BadParameter(click.style("Not authorized for company id {}".format(self.context.get('orgId')), fg='red'))

      pyke.auth.set_cache_token(resp.get('access_token'))
      pyke.auth.set_profile_token(self.context.get('profileName'), resp.get('access_token'))

      return resp

  def print_record_count(self, data):
    d = data['payload']
    if d['totalItems'] > d['currentItemCount']:
      d['startItem'] = (d['page'] - 1) * d['pageSize'] + 1
      d['endItem'] = d['startItem'] - 1 + d['currentItemCount']
      count_str = 'showing {startItem} - {endItem} of {totalItems} total items (page {page} / {totalPages})'
      click.echo(click.style(count_str.format(**d), fg='yellow'))

  def load_users(self):
    resp = self.cli_request('GET', self.build_url('{app}/auth/v1/users'))

    users = {}
    for u in resp['payload']['data']:
      users[u['id']] = u['emailAddress']

    return users

  def print_table(self, data, format, titles={}, handlers={}):
    headers = {}
    stats = {}
    users = {}

    if 'createdBy' in format or 'lastModifiedBy' in format:
      users = self.load_users()

    for c in data:
      child_records = {}
      for k in c:
        if isinstance(c[k], dict):
          for sk in c[k]:
            child_records[k + '__' + sk] = c[k][sk]

      c.update(child_records)

    if len(data) > 0:
      for k in data[0]:
        headers[k] = titles.get(k, k)
    else:
      click.echo('No Records Found')
      return

    for k in headers:
      stats[k] = len(uncolor(headers[k]))

    # TO DO: Cache users and reload when we can't find one
    for c in data:
      for k in c:
        if k in ['createdAt', 'lastModifiedAt']:
          cd = dateutil.parser.parse(c[k])
          c[k] = cd.strftime('%x %X')
        if k in ['createdBy', 'lastModifiedBy']:
          c[k] = users.get(c[k], 'Not Found ({})'.format(c[k]))
        if stats.get(k, 0) < len(str(uncolor(handlers[k](c[k])) if k in handlers else c[k])):
          stats[k] = len(str(uncolor(handlers[k](c[k])) if k in handlers else c[k]))

    for k in headers:
      headers[k] = titles.get(k, k).ljust(stats[k])

    for c in data:
      for k in c:
        c[k] = str(handlers[k](c[k]) if k in handlers else c[k])
        c[k] = c[k] + ' ' * (stats[k] - len(uncolor(c[k])))

    click.echo(format.format(**headers))

    for c in data:
      print(format.format(**c))

  def check_response(self, resp):
    if resp.status_code == 500:
      raise click.ClickException(str(click.style('Unhandled Exception: ', fg='red')) + str(resp.text))
    elif resp.status_code == 404:
      raise click.ClickException(click.style('Not Found', fg='red'))
    elif resp.status_code == 400:
      message = resp.json()['payload']['error']
      if 'apiField' in resp.json()['payload']:
        message += ' (' + resp.json()['payload']['apiField'] + ')'
        raise click.ClickException(click.style(message, fg='red'))

    return resp

  def get_headers(self):
    return {
        'Authorization': 'Bearer ' + self.login_token(),
        'Content-Type': 'application/json'
      }

  def build_url(self, format, extra_data = {}):
    data = pyke.auth.sanatize_data(self.context)
    env = copy.deepcopy(data)
    env.update(self.context)
    env.update(extra_data)
    return format.format(**env).replace(" ", "+")

  def cli_request(self, method, url, data=None, json=None, headers=None, params=None, external=False):
    req_args = self.build_req_args(data, json, headers, params, external)
    req = self.get_request_func(method)
    url = url.replace(" ", "")
    if data is not None:
      resp = self.check_response(req(url, data=data, **req_args, verify=verify_tls))
    else:
      resp = self.check_response(req(url, **req_args, verify=verify_tls))

    if resp.status_code != 200 and external is False:
      self.login_token(renew=True)
      req_args = self.build_req_args(data, json, headers, params, external)
      resp = self.check_response(req(url, **req_args, verify=verify_tls))
      if resp.status_code != 200:
        raise click.BadParameter(click.style('Invalid Credentials', fg='red'))

    if not external:
      return resp.json()
    else:
      return resp

  def build_req_args(self, data, json_data, headers, params, external=False):
    if external is False:
      req_headers = self.get_headers()
      if headers is not None:
        req_headers.update(headers)
    elif headers is not None:
      req_headers = headers
    else:
      req_headers = None

    req_args = {}
    if json_data is not None:
      req_args['json'] = json_data
    if req_headers is not None:
      req_args['headers'] = req_headers
    if params is not None:
      req_args['params'] = params

    return req_args

  def get_request_func(self, method):
    method = method.lower()
    methods = {
      'get': requests.get,
      'head': requests.head,
      'post': requests.post,
      'put': requests.put,
      'patch': requests.patch,
      'delete': requests.delete
    }

    return methods[method]

  def list_companies(self, env, print_table=True):
    conf = pyke.auth.load_config()

    handlers = {
        'instanceType': lambda x: 'dev' if x == 'cc' else x,
    }

    headers = {
      'Authorization': 'Bearer ' + self.login_without_context(env),
      'Content-Type': 'application/json'
    }

    app_url = conf['envs'][env]['app'].replace(" ", "")
    res = requests.get('{}/auth/v1/me/companies?sort=name'.format(app_url), headers=headers, verify=verify_tls)
    resp = res.json()['payload']['data']

    titles = {
      'id': 'id',
      'name': 'Org Name',
      'instanceType': 'Org Type',
      'isTest': 'Test Org'
    }

    if print_table is True:
      self.print_table(resp, '{id} {name} {instanceType} {isTest}', handlers=handlers, titles=titles)

    return resp

  def get_company_id_from_resp(self, id, resp):
    try:
      company_id = next(int(rec.get('id')) for rec in resp if rec.get('name').lower() == id.lower())
      return company_id
    except:
      raise click.ClickException(click.style("Company not found. You can only share/unshare with companies you have provisioned.", fg='red'))

  def get_company_id(self, name_or_id, resp):
    try:
      company_id = int(name_or_id)
      return company_id
    except:
      return self.get_company_id_from_resp(name_or_id, resp)

  def get_company_name(self, id, resp):
    return next(rec.get('name') for rec in resp if int(rec.get('id')) == id)

  def upload_ephemeral(self, filename, content_type):
    resp = self.cli_request('GET', self.build_url('{app}/iot/v1/files/ephemeral?contentType=' + urllib.parse.quote(content_type)))
    if '/' in filename:
      file_name = filename.rsplit('/', 1)[-1]
    else:
      file_name = filename

    ephemeral_url = resp['payload']['data']['url']
    ephemeral_data = resp['payload']['data']['fields']
    ephemeral_data['key'] = ephemeral_data['key'].replace('${filename}', file_name)
    ephemeral_data['Content-Type'] = content_type

    post_res = requests.post(
      ephemeral_url,
      data=ephemeral_data,
      files={'file': open(filename, 'rb')},
      verify=verify_tls )

    return ephemeral_data['key']

  def get_current_access_records(self, object, object_id):
    path = '{}s/{}/access'.format(object, object_id)
    resp = self.cli_request('GET', self.build_url('{app}/iot/v1/' + path))
    return resp['payload']['data']

  def lookup_object_id(self, object_type, id, abort_if_null=True):
    if object_type == 'component-set':
      try:
        object_filter = int(id)
        filter_type = 'id'
      except:
        object_filter = id
        filter_type = 'urlRef'

      resp = self.cli_request('GET',
          self.build_url('{app}/iot/v1/component-sets?{filter_type}={object_filter}',
          extra_data={'filter_type': filter_type, 'object_filter': object_filter}))


      if len(resp.get('payload').get('data', [])) > 0:
        return resp['payload']['data'][0]['id']

      if abort_if_null:
        click.echo(click.style('{} not found'.format(object_type), fg='red'))
        sys.exit()

    else:
      try:
        object_filter = int(id)
        filter_type = 'id'
      except:
        object_filter = id
        filter_type = 'urlRef'

      resp = self.cli_request('GET',
          self.build_url('{app}/iot/v1/containers?{filter_type}={object_filter}',
          extra_data={'filter_type': filter_type, 'object_filter': object_filter}))

      if len(resp.get('payload').get('data', [])) > 0:
        return resp['payload']['data'][0]['id']

      if abort_if_null:
        click.echo(click.style('{} not found'.format(object_type), fg='red'))
        sys.exit()

  def save_config(self, data):
    with open(self.get_config_file(), 'w+') as outfile:
      json.dump(data, outfile, indent=4, sort_keys=True)

  def get_size(self, path, image=True):
    stats = os.stat(path)
    byte_size = stats.st_size
    if byte_size < 1048576:
      return '{} KB'.format(round(byte_size / 1024, 2))
    elif byte_size < 1073741824:
      return '{} MB'.format(round(byte_size / 1048576, 2))
    else:
      if image is True:
        click.echo('Hey! Thanks for being an awesome Lumavate customer.')
        click.echo('While you are waiting for this image to upload to the platform you should check this out!')
        click.echo('https://hackernoon.com/tips-to-reduce-docker-image-sizes-876095da3b34')
        return '{} GB'.format(round(byte_size / 1073741824, 2))

  def parse_version_number(self, version_number):
    try:
      major, minor, patch = version_number.split('.')
      for x in [major, minor, patch]:
        if x is None:
          raise click.ClickException(click.style("Version numbers must be in the format '<major: int>.<minor: int>.<patch: int>' ", fg='red'))
        if int(x) < 0:
          raise click.ClickException(click.style("Version numbers must be in the format '<major: int>.<minor: int>.<patch: int>' ", fg='red'))

      return major, minor, patch
    except:
      raise click.ClickException(click.style("Version numbers must be in the format '<major: int>.<minor: int>.<patch: int>' ", fg='red'))

  def get_config_file(self):
    config_path = str(Path.home()) + '/.luma_cli_config'
    if not os.path.exists(config_path):
      with open(config_path, 'w') as config:
        json.dump({ "envs": {}, "profiles": {} }, config)

    return config_path

  def get_version_id(self, object_type, parent_id, version):
    try:
      version_id = int(version)
    except:
      _version = self.get_version_from_version_number(object_type, parent_id, version)
      if _version is not None:
        version_id = _version.get('id')

    return version_id

  # TO DO:Create logic here to better target smaller page to find version
  def get_versions_from_mask(self, object_type, parent_id, version):
    if object_type == 'component-set':
      versions = self.cli_request('GET', self.build_url('{app}/iot/v1/component-sets/{parent_id}/versions?pagesize=500',\
          {'parent_id': parent_id}))['payload']['data']
    else:
      versions = self.cli_request('GET', self.build_url('{app}/iot/v1/containers/{parent_id}/versions?pagesize=500',\
          {'parent_id': parent_id}))['payload']['data']

    major, minor, patch = version.split('.')

    return sorted([x for x in versions if
        (str(x['major']) == major or major == '*') and
        (str(x['minor']) == minor or minor == '*') and
        (str(x['patch']) == patch or patch == '*')], key=lambda x: (x['major'], x['minor'], x['patch']), reverse=True)

  def get_version_from_version_number(self, object_type, parent_id, version_number):
    if '*' in version_number:
      version = self.get_versions_from_mask(object_type, parent_id, version_number)

    else:
      try:
        major, minor, patch = version_number.split('.')
      except:
        raise click.ClickException(click.style('Version Not Found - Version Number: {}'.format(version_number), fg='red'))

      if object_type == 'component-set':
        version = self.cli_request('GET', self.build_url('{app}/iot/v1/component-sets/{parent_id}/versions?major={major}&minor={minor}&patch={patch}',\
            { 'parent_id': parent_id, 'major': major, 'minor': minor, 'patch': patch }))['payload']['data']
      else:
        version = self.cli_request('GET', self.build_url('{app}/iot/v1/containers/{parent_id}/versions?major={major}&minor={minor}&patch={patch}',\
            { 'parent_id': parent_id, 'major': major, 'minor': minor, 'patch': patch }))['payload']['data']

    if version is not None and len(version) > 0:
      return version[0]

    else:
      raise click.ClickException(click.style('Version Not Found', fg='red'))

  def get_container(self, container):
    try:
      container_id = int(container)
      return self.get_container_by_id(container_id)
    except:
      return self.get_container_by_url_ref(container)

  def get_container_by_url_ref(self, url_ref):
    container_res = self.cli_request('GET', self.build_url('{app}/iot/v1/containers?urlRef={urlRef}', {'urlRef': url_ref}))
    container_list = container_res.get('payload', {}).get('data', {})

    if not container_list or len(container_list) < 1:
      raise click.ClickException(click.style('Container Not Found', fg='red'))

    return container_list[0]

  def get_container_version(self, container_id, version):
    try:
      version_id = int(version)
      return self.get_container_version_by_id(container_id, version_id)
    except:
      return self.get_version_from_version_number('container', container_id, version)

  def get_container_by_id(self, container_id):
    version = self.cli_request('GET', self.build_url('{app}/iot/v1/containers/{container_id}',\
        { 'container_id': container_id}))['payload']['data']

    return version

  def get_container_version_by_id(self, container_id, version_id):
    version = self.cli_request('GET', self.build_url('{app}/iot/v1/containers/{container_id}/versions/{version_id}',\
        { 'container_id': container_id, 'version_id': version_id}))['payload']['data']

    return version

  def get_version(self, object_type, parent_id, version):
    versions = self.get_versions_from_mask(object_type, parent_id, version)
    if versions:
      return versions[0]

    return None

  def resolve_default_icon_path(self):
    pwd = os.path.dirname(os.path.realpath(__file__))
    if platform.system() == 'Windows':
      icon_path = pwd.rsplit('\\', 1)[0]
      return '{}\\images\\default-icon.svg'.format(icon_path)
    else:
      icon_path = pwd.rsplit('/', 1)[0]
      return '{}/images/default-icon.svg'.format(icon_path)

  def show_progress(self, status_id, label=None):
    progress = 0
    error = None
    status_resp = None

    last_progress = 0
    with click.progressbar(length=100, label=label) as bar:
      while progress < 100 and error is None:
        status_resp = self.cli_request('GET', self.build_url('{app}/iot/v1/statuses/{status_id}', {'status_id': status_id}))
        status_data = status_resp.get('payload', {}).get('data', {})
        progress = status_data.get('overallPercent')
        error = status_data.get('errorMessage')
        bar.update(progress - last_progress)
        last_progress = progress

    return status_resp



"""
    These sub-commands and their corresponding logic are both implamented in this file.
    TO DO: Break out the logic for these commands into a seperate behavior file.
"""

from pathlib import Path
from cli import pyke
import requests
import urllib3
import click
import json
import sys
import os

verify_tls = os.environ.get('LUMA_CLI_VERIFY_TLS', None)
if verify_tls == 'false' or verify_tls == 'False':
  verify_tls = False
  urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
else:
  verify_tls = None

@click.group(help="Profiles add company context to commands. Almost all calls to Lumavate require context. \
Profiles use Envs to get tokens and then give those tokens context using the Org ID.")
def profile():
    pass

@click.command('ls', help="List the profiles you have configured.")
@click.option('--format', '-f', default='{profileName} {env} {orgName} {orgId}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--json', 'json_flag', is_flag=True, help='Return raw json.')
def list(format, json_flag):
  config = pyke.auth.load_config()

  data = []
  for key in config['profiles'].keys():
    config['profiles'][key]['profileName'] = str(key)
    data.append(config['profiles'][key])

  if json_flag:
    click.echo(json.dumps(config))
    return

  pyke.Util().print_table(data, format)

@click.command()
@click.option('--profile-name', prompt=True, help='Profile Name')
@click.option('--env', default=None, help='The name of the Env you want to use with this profle')
@click.option('--format', '-f', default='{env} {orgName} {orgId}', help='The available column names for this command are {env} {orgName} {orgId}')
def add(profile_name, env, format):
  if not profile_name.strip():
    raise click.BadParameter(click.style("The --profile-name must contain more than just whitespace.", fg='red'))

  click.echo(' ')
  config = pyke.auth.load_config()

  titles = {
      'envName': 'Env Name',
      'app': 'App',
      'audience': 'Audience',
      'token': 'Token'
  }

  headers = '{envName} {app} {audience} {token}'

  if env is None:
    data = []
    for key in config['envs'].keys():
      config['envs'][key]['envName'] = str(key)
      data.append(config['envs'][key])

    pyke.Util().print_table(data, headers, titles=titles)
    click.echo(' ')

    env = input("Name of the Env you want to use with this profle: ")
    click.echo(' ')

  env_names = config["envs"].keys()
  if env not in env_names:
    raise click.BadParameter(click.style("You must select and Env that you have configured", fg='red'))

  config['profiles'][profile_name] = {
                                "accessToken": " ",
                                "env": env,
                                "orgId": None,
                                "orgName": ""
                            }

  profile_data = config['profiles'][profile_name]

  resp = pyke.Util().list_companies(env)
  click.echo(' ')

  org_id = input("Org ID you want to associate with this profile: ")
  click.echo(' ')

  try:
    org_id = int(org_id)
  except Exception as e:
    raise click.BadParameter(click.style('You must provide the id of the organization', fg='red'))

  if str(org_id) not in [rec.get('id').replace(" ", "") for rec in resp]:
    raise click.BadParameter(click.style('You must provide the id of the organization you have access to', fg='red'))

  org_name = next(rec.get('name') for rec in resp if int(rec.get('id')) == org_id)

  profile_data["orgId"] = org_id
  profile_data["orgName"] = org_name.rstrip()

  config['profiles'][profile_name] = profile_data

  pyke.auth.save_config(config)

  profile_data["profile"] = profile_name
  company_context = get_company_context(profile_name).get('company')

  if company_context:
    config_data = pyke.auth.load_config()
    profile_data = config_data['profiles'][profile_name]
    profile_data['experienceCloudUri'] = company_context['experienceCloudUri']

    pyke.auth.save_config(config_data)

  else:
    raise click.ClickException(click.style('Error getting company context'), fg='red')

  titles = {
      'profile': 'Profile',
      'env': 'Environment',
      'orgName': 'Org Name',
      'orgId': 'Org ID'
  }

  pyke.Util().print_table([profile_data], format, titles=titles)
  click.echo(' ')

def get_company_context(profile):
  config_data = pyke.auth.load_config()
  selected_profile = config_data.get('profiles', {}).get(profile)
  context = config_data.get('envs', {}).get(selected_profile.get('env'))
  context['profileName'] = profile
  context['orgId'] = selected_profile.get('orgId')


  util = pyke.Util(context=context, profile=profile)
  token_resp = util.login()
  headers = {
        'Authorization': 'Bearer ' + token_resp.get('access_token'),
        'Content-Type': 'application/json'
      }

  resp = requests.get(util.build_url('{app}/auth/v1/context'), headers=headers, verify=verify_tls)

  if resp.status_code == 200:
    return resp.json()['payload']['data']
  else:
    raise click.ClickException(click.style('Error getting company context'), fg='red')

@click.command('refresh-token', help='Refreshes the access token stored in the config.')
@click.option('--profile', '-p', prompt=True, help='Profile to login')
def refresh_token(profile):
  config_data = pyke.auth.load_config()

  selected_profile = config_data.get('profiles', {}).get(profile)
  context = config_data.get('envs', {}).get(selected_profile.get('env'))
  context['profileName'] = profile
  context['orgId'] = selected_profile.get('orgId')

  util = pyke.Util(context=context, profile=profile)
  util.login()

  click.echo("Refreshed '{}' access token.".format(profile))

  # Migration path for old profile data without experienceCloudUri
  if 'experienceCloudUri' not in selected_profile.keys():
    company_context = get_company_context(profile).get('company')

    if company_context:
      config_data = pyke.auth.load_config()
      profile_data = config_data['profiles'][profile]
      profile_data['experienceCloudUri'] = company_context['experienceCloudUri']

      pyke.auth.save_config(config_data)

@click.command('rm')
@click.option('--profile', '-p', prompt=True, help='Profile to delete')
def delete(profile):
  config_data = pyke.auth.load_config()
  profile_data = config_data["profiles"][profile]
  del config_data["profiles"][profile]

  pyke.auth.save_config(config_data)

  click.echo("Deleted profile:")
  click.echo(json.dumps(profile_data))


profile.add_command(list)
profile.add_command(delete)
profile.add_command(add)
profile.add_command(refresh_token)



from datetime import datetime
from pathlib import Path
from cli import pyke
import subprocess
import platform
import click
import json
import os

verify_tls = os.environ.get('LUMA_CLI_VERIFY_TLS', None)
if verify_tls == 'false' or verify_tls == 'False':
  verify_tls = False
else:
  verify_tls = None

class ContainerVersionBehavior:
  def __init__(self, profile=None):
    self.util = pyke.Util(profile=profile)
    self.object_type = 'container'

  def force_update (self, container, version):
    container_id = self.util.lookup_object_id(self.object_type, container)
    version_id = self.util.get_version_id(self.object_type, container, version)

    req_data = {'force': True}
    resp = self.util.cli_request('PUT',
        self.util.build_url('{app}/iot/v1/containers/{container_id}/versions/{version_id}',
          {'container_id': container_id, 'version_id': version_id}), json=req_data)

    click.echo(resp)

  def create_editable_version(self, container_id, from_version, env, format):
    if from_version is None:
      raise click.ClickException(click.style("An editable version must come from an exsisting version", fg='red'))

    version = self.util.get_container_version(container_id, from_version)

    if not version.get('editorPort'):
      raise click.ClickException(click.style("An editable version must come from an exsisting version with an editor port", fg='red'))

    versions = self.util.cli_request('GET', self.util.build_url('{app}/iot/v1/containers/{container_id}/versions?major={major}&minor={minor}&sort=patch+desc',\
          { 'container_id': version['containerId'], 'major': version['major'], 'minor': version['minor'] }))['payload']['data']

    if versions is not None and len(versions) > 0:
      latest_version = versions[0]

    version['patch'] = latest_version['patch'] + 1
    version['isEditable'] = True
    version['label'] = 'dev'

    del version['versionNumber']

    if 'env' not in version.keys():
      version['env'] = {}

    for var in env:
      try:
        env_dict = json.loads(var)
        version['env'] = { **version['env'], **env_dict}
      except Exception as e:
        print("Could not serialize {}".format(var))
        print(e)

    resp = self.util.cli_request('POST', self.util.build_url('{app}/iot/v1/containers/{container_id}/versions',\
      {'container_id': container_id}), data=json.dumps(version))

    # TO DO: Add progress bar after create for loading/validating microservice
    status_id = None

    if 'results' in resp['payload']['data'].keys():
      status_id = resp['payload']['data']['results']['statusId']
    if status_id:
      status_resp = self.util.show_progress(status_id, label='Creating Editable Version')
      error_msg = status_resp.get('payload', {}).get('data', {}).get('errorMessage')
      if error_msg is not None:
        raise pyke.CliException('Create version failed. {}'.format(error_msg))

    handlers = {
        'actualState': lambda x: click.style(x, fg='red') if x not in ['running'] else click.style(x, fg='green')
    }

    click.echo(click.style("Using the --is-editable flag ignores all options other than --format and --env", fg='yellow'))
    click.echo(click.style("The editable version will increment the patch number to the first available", fg='yellow'))

    self.util.print_table([resp['payload']['data']], format, handlers=handlers)

  def download_app_zip(self, container, version, path):
    home = str(Path.home())
    if path is None:
      path = home
    else:
      path = '{}/{}'.format(home, path)

    container_id = self.util.lookup_object_id(self.object_type, container)

    try:
      container = self.util.cli_request('GET',
          self.util.build_url('{app}/iot/v1/containers?id={id}', {'id': container_id} ))['payload']['data']
    except:
      raise click.ClickException('Container not found')

    url_ref = container[0].get('urlRef')
    version_id = self.util.get_version_id(self.object_type, container_id, version)

    resp = self.util.cli_request('POST',
        self.util.build_url('{app}/iot/v1/containers/{container_id}/versions/{version_id}/token',
          {'container_id': container_id, 'version_id': version_id}), json={'type': 'editor'})

    token = resp.get('payload', {}).get('data', {}).get('token')

    service_url = self.util.context.get('experienceCloudUri')

    # TODO: Dynamically get integration cloud
    editor_url = '{}/ic/{}/luma-editor/download/application.zip'.format(service_url, url_ref)

    time = datetime.now().microsecond

    if platform.system() == 'Windows':
      zip_path = '{}\\application.{}.zip'.format(path, time)
    else:
      zip_path = '{}/application.{}.zip'.format(path, time)

    command_list = 'curl -L -f --create-dirs --output {} {} -H '.format(zip_path, editor_url).split()
    command_list.append("Authorization: Bearer {}".format(token))
    if verify_tls is False:
      command_list.extend(['--insecure', '--proxy-insecure'])

    subprocess.run(command_list)
    click.echo("File Location: {}".format(zip_path))

  def tail_logs(self, container, version, tail_number=None):
    container = self.util.get_container(container)
    print(container, flush=True)
    container_id = container.get('id')

    version = self.util.get_container_version(container_id, version)
    version_id = version.get('id')

    url_ref = container.get('urlRef')

    editorPort = version.get('editorPort')
    if not editorPort:
      return self.old_logs(container_id, version_id)

    resp = self.util.cli_request('POST',
        self.util.build_url('{app}/iot/v1/containers/{container_id}/versions/{version_id}/token',
          {'container_id': container_id, 'version_id': version_id}), json={'type': 'editor'})

    token = resp.get('payload', {}).get('data', {}).get('token')
    service_url = self.util.context.get('experienceCloudUri')

    if not tail_number:
      tail_number = 100

    # TODO: Dynamically get integration cloud
    editor_url = '{}/ic/{}/luma-editor/logs?tail={}'.format(service_url, url_ref, tail_number)

    command_list = 'curl {} -H '.format(editor_url).split()
    command_list.append("Authorization: Bearer {}".format(token))
    if verify_tls is False:
      command_list.extend(['--insecure', '--proxy-insecure'])

    subprocess.run(command_list)

  def old_logs(self, container_id, version_id):
    resp = self.util.cli_request('GET', self.util.build_url('{app}/iot/v1/containers/{container_id}/versions/{version_id}/logs',\
      {'container_id': container_id, 'version_id': version_id}))

    for x in resp['payload']['data']:
      click.echo(x)





"""
    This behavior file contains the logic for a subset of commands. Logic specific to
    commands should be implamented in corresponding behavior files.
"""

from pathlib import Path
from cli import pyke
import click
import time
import json
import sys
import os

class ComponentSetBehavior:
    def __init__(self, profile=None, component_set=None, format=None, filter=None, page=None, pagesize=None, json=False):
        self.util = pyke.Util(profile=profile)
        if profile is not None:
            self.context = pyke.auth.load_cache()
        self.object_type = 'component-set'
        self.json = json
        self.profile = profile
        self.component_set = component_set
        self.format = format
        self.filter = filter
        self.page = page
        self.pagesize = pagesize


    def list(self):
        data = {'filter': self.filter, 'page': self.page, 'pagesize': self.pagesize}

        resp = self.util.cli_request('GET',
            self.util.build_url('{app}/iot/v1/component-sets?page={page}&pagesize={pagesize}&{filter}', {**data} ))

        if self.json:
            click.echo(json.dumps(resp))
            return

        self.util.print_table(resp['payload']['data'], self.format)
        self.util.print_record_count(resp)

    def delete(self, component_set):
        component_set_id = self.util.lookup_object_id(self.object_type, component_set)

        resp = self.util.cli_request('DELETE',
            self.util.build_url('{app}/iot/v1/component-sets/{component_set_id}', {'component_set_id': component_set_id}))

        if self.json:
            click.echo(json.dumps(resp))
            return

        self.util.print_table([resp['payload']['data']], self.format)

    def create(self, name, url_ref, icon_file):
        if icon_file is None or icon_file == '':
            icon_file = self.util.resolve_default_icon_path()

        post_data = {
          'name': name,
          'urlRef': url_ref,
          'ephemeralKey': self.util.upload_ephemeral(icon_file, 'image/svg+xml')
        }

        resp = self.util.cli_request('POST', self.util.build_url('{app}/iot/v1/component-sets'), data=json.dumps(post_data))

        if self.json:
            click.echo(json.dumps(resp))
            return

        self.util.print_table([resp['payload']['data']], self.format)

    def update(self, component_set, name, icon_file):
        component_set_id = self.util.lookup_object_id(self.object_type, component_set)
        post_data = {}

        if len(name) > 1:
            post_data['name'] = name

        if len(icon_file) > 1:
            post_data['ephemeralKey'] = self.util.upload_ephemeral(icon_file, 'image/svg+xml')

        resp = self.util.cli_request('PUT',
          self.util.build_url('{app}/iot/v1/component-sets/{component_set_id}', {'component_set_id': component_set_id}),
          data=json.dumps(post_data))

        if self.json:
            click.echo(json.dumps(resp))
            return

        self.util.print_table([resp['payload']['data']], self.format)

    def share(self, component_set, add, rm, absolute, current):
        component_set_id = self.util.lookup_object_id(self.object_type, component_set)

        if current:
            if self.json:
                resp = self.util.cli_request('GET',
                    self.util.build_url('{app}/iot/v1/component-sets/{component_set_id}/access',
                    {'component_set_id': component_set_id}))

                click.echo(resp)
                return
            else:
                resp = self.util.cli_request('GET',
                    self.util.build_url('{app}/auth/v1/company/child-companies'))['payload']['data']

                recs = self.util.get_current_access_records('component-set', component_set_id)
                org_names = [self.util.get_company_name(x.get('granteeCompanyId'), resp) for x in recs if x.get('granteeCompanyId') != int(self.util.context.get('orgId'))]
                click.echo(org_names)

                return

        resp = self.util.cli_request('GET',
            self.util.build_url('{app}/auth/v1/company/child-companies'))['payload']['data']

        handlers = {
          'sharedWith': lambda x: x.get('granteeId') if x is not None else '',
          'unsharedWith': lambda x: x.get('granteeId') if x is not None else ''
        }

        if absolute:
          absolute_ids = [self.util.get_company_id(x, resp) for x in absolute]
          data = {
            "id": component_set_id,
            "granteeCompanyIds": [x for x in absolute_ids]
          }

          access_resp = self.util.cli_request('PUT', self.util.build_url('{app}/iot/v1/component-sets/{component_set_id}/access',\
            {'component_set_id': component_set_id}), data=json.dumps(data))

          if self.json:
            click.echo(json.dumps(access_resp))
            return

          recs = self.util.get_current_access_records('component-set', component_set_id)
          access_resp['payload']['data']['resultingGrantees'] =\
            [self.util.get_company_name(x.get('granteeCompanyId'), resp) for x in recs if x.get('granteeCompanyId') != int(self.util.context.get('orgId'))]

          self.util.print_table([access_resp['payload']['data']], self.format)
          return

        recs = self.util.get_current_access_records('component-set', component_set_id)
        sharees = [x.get('granteeCompanyId') for x in recs if x.get('granteeCompanyId') != int(self.context.get('orgId'))]
        if add:
          add_ids = [self.util.get_company_id(x, resp) for x in add]
          sharees.extend(add_ids)

        if rm:
          rm_ids = [self.util.get_company_id(x, resp) for x in rm]
          sharees = [x for x in sharees if x not in rm_ids]

        data = {
          "id": component_set_id,
          "granteeCompanyIds": sharees
        }

        access_resp = self.util.cli_request('PUT',
            self.util.build_url('{app}/iot/v1/component-sets/{component_set_id}/access', {'component_set_id': component_set_id}), data=json.dumps(data))

        if self.json:
          click.echo(json.dumps(access_resp))
          return

        recs = self.util.get_current_access_records('component-set', component_set_id)

        access_resp['payload']['data']['resultingGrantees'] =\
          [self.util.get_company_name(x.get('granteeCompanyId'), resp) for x in recs if x.get('granteeCompanyId') != int(self.context.get('orgId'))]

        self.util.print_table([access_resp['payload']['data']], self.format)



from setuptools.command.install import install
from setuptools import setup, find_packages
from os.path import abspath, dirname, join
import sys, os, stat, json
from pathlib import Path
import mmap

CURRENT_PYTHON = sys.version_info[:2]
REQUIRED_PYTHON = (3, 5)

# This check and everything above must remain compatible with Python 2.7.
if CURRENT_PYTHON < REQUIRED_PYTHON:
    sys.stderr.write("""
==========================
Unsupported Python version
==========================
This version of Lumavate CLI requires Python {}.{}, but you're trying to
install it on Python {}.{}.
This may be because you are using a version of pip that doesn't
understand the python_requires classifier. Make sure you
are installing with pip3, then try again:
    $ pip3 install .
""".format(*(REQUIRED_PYTHON + CURRENT_PYTHON)))
    sys.exit(1)

cli_path = str(Path.home()) + '/.luma_cli_config'
cli_cache = str(Path.home()) + '/.luma_cache'

if not os.path.exists(cli_path):
    with open(cli_path, 'w+') as config:
        json.dump({ "envs": {}, "profiles": {} }, config)

# Don't worry about overwriting cache
with open(cli_cache, 'w+') as config:
    json.dump({ "user": {} }, config)

os.chmod(cli_path, 0o666)
os.chmod(cli_cache, 0o666)

# Read the version number from version.py
with open(abspath(join(dirname(__file__), 'cli', 'version.py'))) as versionFile:
	__version__ = versionFile.read().strip().replace('__version__ = ', '').replace("'", '')

with open('README.md') as f:
    long_description = f.read()

setup(
    name="luma",
    version=__version__,
    description='A CLI to interact with the Lumavate platform',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Gunnar Norred',
    author_email='g.norred@lumavate.com',
    python_requires='>={}.{}'.format(*REQUIRED_PYTHON),
    data_files=[('lumavate-cli', ['cli/images/default-icon.svg'])],
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    install_requires=[
        'click>=7.0',
        'pycparser==2.18',
        'pyparsing==2.2.0',
        'python-dateutil==2.7.3',
        'requests==2.19.1',
        'docker-py==1.10.6',
        'PyQRCode==1.2.1'
    ],
    entry_points='''
        [console_scripts]
        luma=cli.cli:cli
    '''
)

zshrc = str(Path.home()) + '/.zshrc'
bashrc = str(Path.home()) + '/.bash_profile'

tab_comp = False

if os.path.exists(zshrc):
    with open(zshrc, 'rb', 0) as zsh_config, mmap.mmap(zsh_config.fileno(), 0, access=mmap.ACCESS_READ) as s:
        if s.find(b'_LUMA_COMPLETE=source_zsh luma') != -1:
            tab_comp = True
            print("Completion already activated for ZSH")

    if tab_comp is False:
        with open(zshrc, 'a') as zsh:
            zsh.write('eval "$(_LUMA_COMPLETE=source_zsh luma)"')

if os.path.exists(bashrc):
    with open(bashrc, 'rb', 0) as bash_config, mmap.mmap(bash_config.fileno(), 0, access=mmap.ACCESS_READ) as s:
        if s.find(b'_LUMA_COMPLETE=source luma') != -1:
            tab_comp = True
            print("Completion already activated for Bash")

    if tab_comp is False:
        with open(bashrc, 'a+') as bash:
            bash.write('eval "$(_LUMA_COMPLETE=source luma)"')
else:
    with open(bashrc, 'a+') as bash:
        bash.write('eval "$(_LUMA_COMPLETE=source luma)"')



"""
    These sub-commands and their corresponding logic are both implamented in this file.
    TO DO: Break out the logic for these commands into a seperate behavior file.
"""

from pathlib import Path
from cli import pyke
import requests
import urllib3
import click
import json
import os

verify_tls = os.environ.get('LUMA_CLI_VERIFY_TLS', None)
if verify_tls == 'false' or verify_tls == 'False':
  verify_tls = False
  urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
else:
  verify_tls = None

@click.group('org', help='Get information about Orgs such as parent-child org relationships')
def organization():
  pass

@click.command('ls', help="List the Orgs you can make profiles for. \
The 'isTest' column in the returned table indicates if a studio is a test org or a prod org.")
@click.option('--env', prompt=True, help='The Env to point to')
@click.option('--format', '-f', default='{id} {name} {instanceType} {isTest}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--filter', default='')
@click.option('--json', 'json_flag', is_flag=True, help='Return raw json from the platform.')
def list(env, format, filter, json_flag):
  handlers = {
    'instanceType': lambda x: 'dev' if x == 'cc' else x
  }

  if 'instanceType=dev' in filter:
    filter = filter.replace('instanceType=dev', 'instanceType=cc')

  if 'isTest=None' in filter:
    filter = filter.replace('isTest=None', 'instanceType=cc')

  config = pyke.auth.load_config()

  headers = {
      'Authorization': 'Bearer ' + pyke.Util().login_without_context(env),
      'Content-Type': 'application/json'
  }
  try:
      app_name = config['envs'][env]['app'].replace(" ", "")
  except:
      raise click.ClickException(click.style("Env not configured", fg='red'))

  resp = requests.get('{}/auth/v1/me/companies?sort=name&{}'.format(app_name, filter), headers=headers, verify=verify_tls)

  if json_flag:
      click.echo(resp.text)
      return

  pyke.Util().print_table(resp.json()['payload']['data'], format, handlers=handlers)

@click.command('child-orgs', help='List the Orgs that a parent Org can share objects with')
@click.option('--profile', '-p', prompt=True, help='The org whose children you want to list', autocompletion=pyke.auth.get_profile_names)
@click.option('--format', '-f', default='{id} {name} {instanceType} {isTest}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--filter', default='')
@click.option('--json', 'json_flag', is_flag=True, help='Return raw json from the platform.')
@pyke.auth.profile_context_required
def child_companies(profile, format, filter, json_flag):
  util = pyke.Util()

  if 'instanceType=dev' in filter:
    filter = filter.replace('instanceType=dev', 'instanceType=cc')

  if 'isTest=None' in filter:
      filter = filter.replace('isTest=None', 'instanceType=cc')

  handlers = {
    'instanceType': lambda x: 'dev' if x == 'cc' else x
  }

  resp = util.cli_request('GET',
    util.build_url('{app}/auth/v1/company/child-companies?{filter}', {'filter': filter}))

  if json_flag:
    click.echo(json.dumps(resp))
    return

  util.print_table(resp['payload']['data'], format, handlers=handlers)

@click.command()
@click.option('--name', prompt=True)
@click.option('--type', default='studio', type=click.Choice(['studio', 'dev']), prompt=True)
@click.option('--format', '-f', default='{id} {name} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--json', 'json_flag', is_flag=True, help='Return raw json from the platform.')
def provision(name, type, format, json_flag):
  if type == 'dev':
    type = 'cc'

  resp = check_response(requests.get(build_url('{app}/auth/v1/context'), headers=get_headers(), verify=verify_tls)).json()

  domain_prefix = name.replace(' ', '-').lower()

  data = {
    'name': name,
    'domainPrefix': domain_prefix,
    'experienceCloudId': 1,
    'authRealmId': 1,
    'instanceType': type
  }

  resp = check_response(requests.post(build_url('{app}/auth/v1/admin/companies'), json=data, headers=get_headers(), verify=verify_tls))

  if json_flag:
    click.echo(resp.text)
    return

  print_table([resp.json()['payload']['data']], format)

organization.add_command(list)
organization.add_command(child_companies)

# To Do: Finish this
#organization.add_command(provision)
